#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
UWU, the Uncomplicated Web Uploader

Copyright (C) 2020 Brendon Lloyd Higgins <brendon@quantumfurball.net>.

http://quantumfurball.net/tag/uwu.html

This program is free software: you can redistribute it and/or modify it under the terms of the GNU
General Public License as published by the Free Software Foundation, version 3 of the License.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License along with this program. If not,
see <http://www.gnu.org/licenses/>.
"""

import cgi
import cgitb; cgitb.enable() # Optional, for nice exceptions...
import os
import io
import time
import tempfile
import hashlib
import configparser


# NOTE: This variable must be a path to a writable folder containing the configuration file. Change
# this as appropriate upon deployment.
LANDING = "uwu_landing"

# NOTE: You can set this to False to disable the test server. Ordinarily the test server won't run
# if we detect that we're being run as a CGI script. The Python built-in CGIHTTPServer "cannot
# execute redirects (HTTP code 302), because code 200 (script output follows) is sent prior to
# execution of the CGI script. This pre-empts the status code."
# (https://docs.python.org/2/library/cgihttpserver.html)
# The Javascript in the template is written such that this does not affect things: if the Location
# header is present, it will be acted on. (Without Javascript enabled, however, the page will not refresh.)
TEST_SERVER_ENABLED = True

TEST_SERVER_DEFAULT_PORT = 8000


CONFIG_FILE = os.path.join(LANDING, 'config.ini')

HTML_TEMPLATE = 'tmpl.html'
UPLOADED_TEMPLATE = 'uploaded_tmpl.html'
FILE_NOT_ACCESSIBLE_TEMPLATE = 'file_not_accessible_tmpl.html'
FILE_INCOMPLETE_TEMPLATE = 'file_incomplete_tmpl.html'
INCORRECT_SPACE_OR_KEY_TEMPLATE = 'incorrect_space_or_key_tmpl.html'
EXCEEDS_MAXFILESIZE = 'exceeds_maxfilesize_tmpl.html'
EXCEEDS_MAXFILES = 'exceeds_maxfiles_tmpl.html'
EXCEEDS_CAPACITY = 'exceeds_capacity_tmpl.html'

# Ensure user and group, and no-one else, has RW(X) access.
FOLDER_PERMISSION = 0o770
FILE_PERMISSION = 0o660


class FileNotAccessibleException(Exception):
    pass


class FileIncompleteException(Exception):
    pass


class SpaceFieldMissingException(Exception):
    pass


class KeyFieldMissingException(Exception):
    pass


class IncorrectSpaceOrKeyException(Exception):
    pass


class ExceedsSpaceMaxFileSize(Exception):
    pass


class ExceedsSpaceCapacity(Exception):
    pass


class ExceedsSpaceMaxFiles(Exception):
    pass


def apply_template(template, fields):
    tmpl = open(os.path.join(LANDING, template)).read()
    # str.format() isn't great because its syntax conflicts with CSS. Really, we just want to
    # replace fields.
    for field, value in fields.items():
        tmpl = tmpl.replace('{' + field + '}', str(value))
    return tmpl


def metadata_filename(filename):
    return filename + '.meta'


def is_metadata_filename(filename):
    return filename.endswith('.meta')


def check_space_key(form, config):
    space = form.getfirst('space')
    if space is None:
        raise SpaceFieldMissingException()

    key = form.getfirst('key')
    if key is None:
        raise KeyFieldMissingException()

    try:
        space_config = config[space]
    except KeyError:
        # No such space...
        raise IncorrectSpaceOrKeyException()

    try:
        space_key = space_config['key']
    except KeyError:
        # No key is assigned to this space...
        space_key = ''

    if key != space_key:
        raise IncorrectSpaceOrKeyException()

    return space


def timestamp_str():
    return time.strftime('%Y-%m-%d %H:%M:%S %Z')


def create_folder_if_missing(upload_folder):
    try:
        os.mkdir(upload_folder)
        os.chmod(upload_folder, FOLDER_PERMISSION)
    except OSError:
        if not os.path.isdir(upload_folder):
            raise


def get_space_maxfilesize(config, space):
    return config.getint(space, 'maxfilesize', fallback = -1)


def get_space_capacity(config, space):
    return config.getint(space, 'capacity', fallback = -1)


def get_space_maxfiles(config, space):
    return config.getint(space, 'maxfiles', fallback = -1)


def get_filesize(fileitem):
    try:
        # This only works if there is a real temporary file to query.
        return os.fstat(fileitem.file.fileno()).st_size
    except io.UnsupportedOperation:
        # Sometimes there is actually no real temporary file.
        return len(fileitem.value)


def check_size_is_acceptable(filesize, maxfilesize):
    if maxfilesize >= 0 and filesize > maxfilesize:
        raise ExceedsSpaceMaxFileSize()


def count_and_size_folder_contents(folder):
    # Excludes metadata files.
    # Note: Doesn't recurse into subfolders...
    items = [f for f in os.scandir(folder) if f.is_file() and not is_metadata_filename(f.name)]
    sizes = [f.stat().st_size for f in items]
    return (len(items), sum(sizes))


def create_upload_file(upload_folder):
    # Use tempname, not file name from the form, for security...
    file = tempfile.NamedTemporaryFile(mode = 'wb', prefix = '',
                                       dir = upload_folder, delete = False)
    os.chmod(file.name, FILE_PERMISSION)
    return file


def create_upload_file_and_check_quota(folder, new_size, capacity, maxfiles):
    # Check quotas. Size first, so we don't occupy a count slot unnecessarily.
    (count, size) = count_and_size_folder_contents(folder)
    if capacity >= 0 and size + new_size > capacity:
        raise ExceedsSpaceCapacity()
    
    if maxfiles >= 0 and count >= maxfiles:
        raise ExceedsSpaceMaxFiles()
    
    # Ignore potential racing with other processes. We might be a bit off in some cases, but
    # shouldn't be bad enough to worry.
    return create_upload_file(folder)


def create_metadata_file(space, upload_file, fileitem):
    fmetaname = metadata_filename(upload_file.name)
    fmeta = open(fmetaname, 'w')
    os.chmod(fmetaname, FILE_PERMISSION)
    fmeta.write('''Began processing: {}
Script name: {}
Source IP: {}
Source user agent: {}
Space: {}
Upload ID: {}
Original file name: {}
'''.format(timestamp_str(),
           SCRIPT_NAME,
           os.environ["REMOTE_ADDR"],
           os.environ["HTTP_USER_AGENT"],
           space,
           os.path.basename(upload_file.name),
           fileitem.filename))
    return fmeta


def complete_metadata_file(fmeta, hash_md5):
    fmeta.write('''Hash MD5 sum: {}
Completed processing: {}
'''.format(hash_md5, timestamp_str()))
    fmeta.close()


def copy_and_hash(fin, fout):
    hash_md5 = hashlib.md5()
    while True:
        chunk = fin.read(1024 * 1024)
        if not chunk:
            break
        fout.write(chunk)
        hash_md5.update(chunk)
    return hash_md5.hexdigest()


def upload(form):
    config = configparser.ConfigParser(interpolation = None)
    config.read(CONFIG_FILE)

    try:
        space = check_space_key(form, config)

        fileitem = form['file']
        if not fileitem.file:
            raise FileNotAccessibleException()

        if fileitem.done == -1:
            raise FileIncompleteException()

        maxfilesize = get_space_maxfilesize(config, space)
        filesize = get_filesize(fileitem)
        check_size_is_acceptable(filesize, maxfilesize)

        capacity = get_space_capacity(config, space)
        maxfiles = get_space_maxfiles(config, space)

        upload_folder = os.path.join(LANDING, space)
        create_folder_if_missing(upload_folder)

        fout = create_upload_file_and_check_quota(upload_folder, filesize, capacity, maxfiles)
        fmeta = create_metadata_file(space, fout, fileitem)

        hash_md5 = copy_and_hash(fileitem.file, fout)
        fout.close()

        complete_metadata_file(fmeta, hash_md5)

        # Redirect, to prevent refresh re-uploading files.
        print('Status: 303 See other')
        print('Location: {}?succ_id={}&hash_md5={}'.format(
            SCRIPT_NAME, os.path.basename(fout.name), hash_md5))
        print() # Critical: blank line so the server knows the above is HTTP header information...

        return
    except FileNotAccessibleException:
        result = apply_template(FILE_NOT_ACCESSIBLE_TEMPLATE, {})
    except FileIncompleteException:
        result = apply_template(FILE_INCOMPLETE_TEMPLATE, {})
    except IncorrectSpaceOrKeyException:
        result = apply_template(INCORRECT_SPACE_OR_KEY_TEMPLATE, {})
    except ExceedsSpaceMaxFileSize:
        result = apply_template(EXCEEDS_MAXFILESIZE, {
            'FILESIZE': filesize,
            'MAXFILESIZE': maxfilesize
            })
    except ExceedsSpaceCapacity:
        result = apply_template(EXCEEDS_CAPACITY, {
            'FILESIZE': filesize,
            'CAPACITY': capacity
            })
    except ExceedsSpaceMaxFiles:
        result = apply_template(EXCEEDS_MAXFILES, {
            'MAXFILES': maxfiles
            })

    print('Content-type: text/html')
    print()
    print(apply_template(HTML_TEMPLATE, {
        'SCRIPT_NAME': SCRIPT_NAME,
        'RESULT': result
        }))


def succeed(form):
    result = apply_template(UPLOADED_TEMPLATE, {
        'ID': form.getfirst('succ_id'),
        'HASH_MD5': form.getfirst('hash_md5', 'Missing')
        })

    print('Content-type: text/html')
    print()
    print(apply_template(HTML_TEMPLATE, {
        'SCRIPT_NAME': SCRIPT_NAME,
        'RESULT': result
        }))


def view(form):
    print('Content-type: text/html')
    print()
    print(apply_template(HTML_TEMPLATE, {
        'SCRIPT_NAME': SCRIPT_NAME,
        'RESULT': ''
        }))


def handle_request():
    # If we happen to be running on Windows, stdio needs to be set for binary mode.
    try:
        import msvcrt
        msvcrt.setmode (0, os.O_BINARY) # stdin  = 0
        msvcrt.setmode (1, os.O_BINARY) # stdout = 1
    except ImportError:
        pass

    global SCRIPT_NAME
    form = cgi.FieldStorage()

    SCRIPT_NAME = os.environ['SCRIPT_NAME']
    # When running with the test server, SCRIPT_NAME starts with "//", which breaks things.
    if SCRIPT_NAME.startswith("//"):
        SCRIPT_NAME = SCRIPT_NAME[1:]

    if 'succ_id' in form:
        succeed(form)
    elif 'file' in form:
        upload(form)
    else:
        view(form)


def run_test_server():
    import http.server
    import sys

    server = http.server.HTTPServer
    handler = http.server.CGIHTTPRequestHandler

    if len(sys.argv) > 2:
        print("Accepts only one optional argument: the port to open for the test server.")
        exit(1)
    elif len(sys.argv) == 2:
        server_port = int(sys.argv[1])
    else:
        server_port = TEST_SERVER_DEFAULT_PORT
    server_address = ("", server_port)

    # NOTE This causes the SCRIPT_NAME environment variable to begin with an extra /. 'Tis the
    # nature of the Python CGI test server, unfortunately.
    handler.cgi_directories = ["/"]

    httpd = server(server_address, handler)

    print("Starting server. CTRL-C to terminate.")
    httpd.serve_forever()


# Main. Detect if we're being run as a CGI script or not, and if not, start a test server.
if TEST_SERVER_ENABLED and "SERVER_SOFTWARE" not in os.environ:
    run_test_server()
else:
    handle_request()
